/***************************************************************************

  CFont.cpp

  (c) 2000-2003 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CFONT_CPP


#include "gambas.h"

#include <qapplication.h>
#include <qstringlist.h>
#include <qfontmetrics.h>

#include "CWidget.h"
#include "CDraw.h"
#include "CFont.h"

void *CLASS_Font;

static CFONT font_font;


static void CFONT_ref(CFONT *font)
{
  if (font->font == CFONT_WIDGET)
  {
    //printf("CFONT_ref %p\n", font);
    GB.Ref(CFONT_TO_CWIDGET(font));
  }
}

static void CFONT_unref(CFONT **font, int can_free)
{
  CWIDGET *wid;

  if ((*font)->font == CFONT_WIDGET)
  {
    //printf("CFONT_unref %p %d\n", *font, can_free);

    wid = CFONT_TO_CWIDGET(*font);
    GB.UnrefKeep((void **)&wid, can_free);
    if (wid == 0)
      *font = 0;
  }
}

static int CFONT_check(CFONT *font)
{
  //qDebug("CFONT_check %p", font->font);

  if (font->font == CFONT_WIDGET)
    return CWIDGET_check(CFONT_TO_CWIDGET(font));
  else
    return 0;
}

QFont CFONT_get_font(CFONT *font)
{
  if (font == 0)
    return qApp->font();
  else if (font->font == CFONT_WIDGET)
    return CFONT_TO_CWIDGET(font)->widget->font();
  else if (font->font == CFONT_SCREEN)
    return qApp->font();
  else if (font->font == CFONT_DEVICE)
    return DRAW_get_font();
  else
    return *(font->font);
}

static void CFONT_manage(int prop, CFONT *font, void *_param)
{
  QFont qfont; // $$ DOUTEUX ! Devrait tre allou sur le tas !!

  qfont = CFONT_get_font(font);

  if (READ_PROPERTY)
  {
    switch(prop)
    {
      case CFONT::Name: GB.ReturnNewZeroString(qfont.family()); break;
      case CFONT::Size: GB.ReturnInteger(qfont.pointSize()); break;
      case CFONT::Bold: GB.ReturnBoolean(qfont.bold()); break;
      case CFONT::Italic: GB.ReturnBoolean(qfont.italic()); break;
      case CFONT::Underline: GB.ReturnBoolean(qfont.underline()); break;
      case CFONT::StrikeOut: GB.ReturnBoolean(qfont.strikeOut()); break;
    }
  }
  else
  {
    switch (prop)
    {
      case CFONT::Name: qfont.setFamily(GB.ToZeroString(PROP(GB_STRING))); break;
      case CFONT::Size: qfont.setPointSize(VPROP(GB_INTEGER)); break;
      case CFONT::Bold: qfont.setBold(VPROP(GB_BOOLEAN)); break;
      case CFONT::Italic: qfont.setItalic(VPROP(GB_BOOLEAN)); break;
      case CFONT::Underline: qfont.setUnderline(VPROP(GB_BOOLEAN)); break;
      case CFONT::StrikeOut: qfont.setStrikeOut(VPROP(GB_BOOLEAN)); break;
    }

    if (CFONT_is_widget(font))
      CFONT_TO_CWIDGET(font)->widget->setFont(qfont);
    else if (CFONT_is_screen(font))
      qApp->setFont(qfont);
    else if (CFONT_is_device(font))
      DRAW_set_font(qfont);
    else
      *(font->font) = qfont;
  }
}


BEGIN_METHOD_VOID(CFONT_init)

  CLASS_Font = GB.FindClass("Font");

  font_font.klass = CLASS_Font;
  font_font.font = 0;

END_METHOD


BEGIN_PROPERTY(CFONT_name)

  CFONT_manage(CFONT::Name, OBJECT(CFONT), _param);

END_PROPERTY

BEGIN_PROPERTY(CFONT_size)

  CFONT_manage(CFONT::Size, OBJECT(CFONT), _param);

END_PROPERTY

BEGIN_PROPERTY(CFONT_bold)

  CFONT_manage(CFONT::Bold, OBJECT(CFONT), _param);

END_PROPERTY

BEGIN_PROPERTY(CFONT_italic)

  CFONT_manage(CFONT::Italic, OBJECT(CFONT), _param);

END_PROPERTY

BEGIN_PROPERTY(CFONT_underline)

  CFONT_manage(CFONT::Underline, OBJECT(CFONT), _param);

END_PROPERTY

BEGIN_PROPERTY(CFONT_strikeout)

  CFONT_manage(CFONT::StrikeOut, OBJECT(CFONT), _param);

END_PROPERTY


static void add(QString &str, const QString& data)
{
  if (str.length())
    str += ',';

  str += data;
}

BEGIN_METHOD_VOID(CFONT_to_string)

  QFont qfont; // $$ DOUTEUX ! Devrait tre allou sur le tas !!
  QString str;

  qfont = CFONT_get_font(OBJECT(CFONT));

  //str = qfont.family().left(1).upper() + qfont.family().mid(1).lower() + " " + QString::number(qfont.pointSize());
  add(str, qfont.family());
  add(str, QString::number(qfont.pointSize()));
  if (qfont.bold())
    add(str, "Bold");
  if (qfont.italic())
    add(str, "Italic");
  if (qfont.underline())
    add(str, "Underline");
  if (qfont.strikeOut())
    add(str, "StrikeOut");

  GB.ReturnNewZeroString(str.latin1());

END_METHOD


BEGIN_METHOD(CFONT_get, GB_STRING str)

  QString str(GB.ToZeroString(ARG(str)));
  QStringList list;
  QString name, elt, flag;
  ushort size = 0;
  bool number;

  // (!) Remove this test later, it is for backward compatibility

  list = QStringList::split(str.find(',') < 0 ? " " : ",", str);

  if (font_font.font)
    delete font_font.font;

  font_font.font = new QFont(qApp->font());
  font_font.font->setBold(false);
  font_font.font->setItalic(false);
  font_font.font->setUnderline(false);
  font_font.font->setStrikeOut(false);

  for (QStringList::Iterator it = list.begin(); it != list.end(); ++it )
  {
    elt = (*it);
    flag = elt.upper();
    size = elt.toUShort(&number);

    if (flag == "BOLD")
      font_font.font->setBold(true);
    else if (flag == "ITALIC")
      font_font.font->setItalic(true);
    else if (flag == "UNDERLINE")
      font_font.font->setUnderline(true);
    else if (flag == "STRIKEOUT")
      font_font.font->setStrikeOut(true);
    else if (number && size > 0)
      font_font.font->setPointSize(size);
    else
    {
      if (!name.isEmpty())
        name += ' ';
      name += elt;
    }
  }

  //qDebug("name = %s", name.latin1());
  if (name.length())
    font_font.font->setFamily(name);

  GB.ReturnObject(&font_font);

END_METHOD


BEGIN_METHOD(CFONT_width, GB_STRING text)

  QFontMetrics fm(CFONT_get_font(THIS));
  QStringList sl;
  int w, width = 0;
  int i;

  QString str;

  sl = QStringList::split('\n', QSTRING_ARG(text));

  for (i = 0; i < (int)sl.count(); i++)
  {
    w = fm.width(sl[i]);
    if (w > width) width = w;
  }

  GB.ReturnInteger(width);

END_METHOD


BEGIN_PROPERTY(CFONT_ascent)

  QFontMetrics fm(CFONT_get_font(THIS));

  GB.ReturnInteger(fm.ascent());

END_PROPERTY


BEGIN_PROPERTY(CFONT_descent)

  QFontMetrics fm(CFONT_get_font(THIS));

  GB.ReturnInteger(fm.descent());

END_PROPERTY


BEGIN_METHOD(CFONT_height, GB_STRING text)

  QFontMetrics fm(CFONT_get_font(THIS));
  QString s;

  s = QSTRING_ARG(text);

  GB.ReturnInteger(fm.height() * (1 + s.contains('\n')));

END_METHOD





GB_DESC CFontDesc[] =
{
  GB_DECLARE("Font", 0), GB_VIRTUAL_CLASS(),
  GB_HOOK_REF(CFONT_ref), GB_HOOK_UNREF(CFONT_unref),
  GB_HOOK_CHECK(CFONT_check),

  GB_STATIC_METHOD("_init", NULL, CFONT_init, NULL),

  GB_PROPERTY("Name", "s", CFONT_name),
  GB_PROPERTY("Size", "i", CFONT_size),
  GB_PROPERTY("Bold", "b", CFONT_bold),
  GB_PROPERTY("Italic", "b", CFONT_italic),
  GB_PROPERTY("Underline", "b", CFONT_underline),
  GB_PROPERTY("StrikeOut", "b", CFONT_strikeout),

  GB_METHOD("ToString", "s", CFONT_to_string, NULL),

  GB_METHOD("Width", "i", CFONT_width, "(Text)s"),
  GB_METHOD("Height", "i", CFONT_height, "(Text)s"),

  GB_STATIC_METHOD("_get", "Font", CFONT_get, "s"),

  GB_PROPERTY_READ("Ascent", "i", CFONT_ascent),
  GB_PROPERTY_READ("Descent", "i", CFONT_descent),

  GB_END_DECLARE
};


